/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "FilterSelector.h"

#include <Modules/Mavlink/MavlinkVersionHeader.h>
#include <Modules/SkywardHubStrings.h>
#include <qglobal.h>
#include <qregularexpression.h>

#include <QDebug>
#include <QLabel>
#include <QPushButton>

void FilterSelector::setupUi()
{
    setAttribute(Qt::WA_DeleteOnClose, true);
    setWindowTitle("Build Filter");

    QVBoxLayout *layout = new QVBoxLayout;

    systemsComboBox = new QComboBox;
    systemsComboBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed));
    systemsComboBox->addItem("Main");
    systemsComboBox->addItem("Payload");
    systemsComboBox->addItem("Rig");
    systemsComboBox->addItem("Gs Receiver");
    layout->addWidget(systemsComboBox);

    messagesComboBox = new QComboBox;
    messagesComboBox->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    messagesComboBox->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
    messagesComboBox->addItems(messages.keys());
    layout->addWidget(messagesComboBox);

    fieldsWidget = new QListWidget;
    layout->addWidget(fieldsWidget);

    QHBoxLayout *buttons = new QHBoxLayout;
    QPushButton *cancel  = new QPushButton("Cancel");
    buttons->addWidget(cancel);
    QPushButton *select = new QPushButton("Select");
    buttons->addWidget(select);
    layout->addLayout(buttons);

    // If the filter is not empty, parse the filter and set system and message
    if (filter.getTopic().toString() != Topic().toString())
    {
        QRegularExpression filterPattern("^((?:\\w|_)+)/In/((?:\\w|_)+)");
        auto match = filterPattern.match(filter.getTopic().toString());

        if (match.hasMatch())
        {
            auto system    = match.captured(1);
            currentMessage = match.captured(2);

            systemsComboBox->setCurrentText(system);
        }
    }
    else
    {
        currentMessage = messages.keys().at(0);
    }
    messagesComboBox->setCurrentText(currentMessage);

    // Update the fields list with the current selected message
    setTopic(currentMessage);

    connect(messagesComboBox, &QComboBox::currentTextChanged, this,
            [=](QString key)
            {
                if (messages.contains(key))
                {
                    filter = getFilter();
                    setTopic(key);
                }
            });
    connect(cancel, &QPushButton::clicked, [this]() { deleteLater(); });
    connect(select, &QPushButton::clicked,
            [this]()
            {
                filter = getFilter();
                emit filterSelected(filter);
                deleteLater();
            });

    setLayout(layout);
}

void FilterSelector::selectFilter(FilterSelector::Handler handler)
{
    FilterSelector *selector = new FilterSelector();
    QObject::connect(selector, &FilterSelector::filterSelected, handler);
    selector->show();
}

void FilterSelector::selectFilter(const Filter &filter,
                                  FilterSelector::Handler handler)
{
    FilterSelector *selector = new FilterSelector(filter);
    QObject::connect(selector, &FilterSelector::filterSelected, handler);
    selector->show();
}

FilterSelector::FilterSelector()
{
    parseMessagesList();
    setupUi();
}

FilterSelector::FilterSelector(const Filter &filter) : filter(filter)
{
    parseMessagesList();
    setupUi();
}

void FilterSelector::parseMessagesList()
{
    mavlink_message_info_t messagesList[256] = MAVLINK_MESSAGE_INFO;

    for (unsigned int i = 0; i < 255; i++)
    {
        // Check if the message is empty
        if (strcmp(messagesList[i].name, "EMPTY") != 0)
        {
            QList<QString> fields;

            for (unsigned int ii = 0; ii < messagesList[i].num_fields; ii++)
                fields.append(messagesList[i].fields[ii].name);

            fields.append(SkywardHubStrings::mavlink_component_id_name);
            fields.append(SkywardHubStrings::mavlink_system_id_name);

            messages[messagesList[i].name] = fields;
        }
    }
}

void FilterSelector::setTopic(QString topic)
{
    // Clear the fields
    fieldsWidget->clear();

    // Add the new topic fields
    fieldsWidget->addItems(messages[topic]);
    for (int i = 0; i < fieldsWidget->count(); i++)
    {
        auto item = fieldsWidget->item(i);
        item->setFlags(item->flags() | Qt::ItemIsUserCheckable);

        // Check the item if it was checked before
        if (filter.getFields().contains(item->text()))
            item->setCheckState(Qt::Checked);
        else
            item->setCheckState(Qt::Unchecked);
    }

    currentMessage = topic;
}

Filter FilterSelector::getFilter()
{
    auto targetSystem = systemsComboBox->currentText();
    auto messageName  = messagesComboBox->currentText();

    auto filter = Filter(Topic(targetSystem + "/In/" + messageName));

    for (int i = 0; i < fieldsWidget->count(); i++)
    {
        auto item = fieldsWidget->item(i);
        if (item->checkState())
            filter.addField(item->text());
    }

    return filter;
}
