/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "BaseMavlinkModule.h"

#include <Modules/SkywardHubStrings.h>
#include <qcombobox.h>
#include <qglobal.h>

#include <utility>

#include "Core/Message/Topic.h"

int BaseMavlinkModule::ACTIVE_MAVLINK_MODULES = 0;

BaseMavlinkModule::BaseMavlinkModule(MavlinkPort *port, ModuleId id)
    : Module(id), port(port), mavlinkCodec(port, this)
{
    setupUi();

    connect(&mavlinkCodec, &MavlinkCodec::msgReceived, this,
            &BaseMavlinkModule::onMsgReceived);
    connect(&linkQualityTimer, &QTimer::timeout, this,
            &BaseMavlinkModule::onLinkQualityTimerTick);
}

BaseMavlinkModule::~BaseMavlinkModule()
{
    if (startToggleButton->state())
        stop();

    MessageBroker::getInstance().unsubscribe(
        Filter::fromString(SkywardHubStrings::commandsTopic + "/*"), this);
}

XmlObject BaseMavlinkModule::toXmlObject()
{
    XmlObject obj = childToXmlObject();
    obj.addAttribute("log_file", logCheckBox->isChecked() ? 1 : 0);

    return obj;
}

void BaseMavlinkModule::fromXmlObject(const XmlObject &obj)
{
    childFromXmlObject(obj);

    int logFile;
    obj.getAttribute("log_file", logFile);
    logCheckBox->setChecked(logFile != 0);
}

void BaseMavlinkModule::onMsgReceived(const Message &msg)
{
    msgArrived++;

    // The target system should always be selected when receiving a message
    if (!targetSystem.has_value())
    {
        error("Mavlink",
              "No target system selected while receiving a message.");
    }

    // Set the topic with the selected target system
    auto msgFixed = msg;
    msgFixed.setTopic(
        Topic(targetSystem.value().first + "/In/" + msg.getTopic().toString()));

    MessageBroker::getInstance().publish(msgFixed);
}

void BaseMavlinkModule::onLinkQualityTimerTick()
{
    float ratio = (float)msgArrived / linkQualityPeriod * 1000.0;
    msgArrived  = 0;
    rateLabel->setText("Rate: " + QString::number(ratio) + " msg/s");
}

void BaseMavlinkModule::onStartStreamToggled(bool state)
{
    if (state)
        onStartClicked();
    else
        onStopClicked();
}

void BaseMavlinkModule::onTargetSysChanged(int itemIndex)
{
    QString sysName           = sysIdComboBox->itemText(itemIndex);
    MavlinkCodec::SysId sysId = static_cast<MavlinkCodec::SysId>(
        sysIdComboBox->itemData(itemIndex).toInt());

    // Unsubscribe from previous system
    if (targetSystem.has_value())
    {
        auto oldFilter =
            Filter::fromString(targetSystem.value().first + "/Out/*");
        MessageBroker::getInstance().unsubscribe(oldFilter, this);
        qDebug() << "Unsubscribed from " << oldFilter.toString();
    }

    // Subscribe to new system
    auto newFilter = Filter::fromString(sysName + "/Out/*");
    MessageBroker::getInstance().subscribe(
        newFilter, this,
        [this](const Message &message, const Filter &filter)
        { onCommandReceived(message); });
    qDebug() << "Subscribed to " << newFilter.toString();

    // Update the target system
    targetSystem = std::pair(sysName, sysId);
}

void BaseMavlinkModule::onStartClicked()
{
    if (open())
    {
        // Keep track of all active and listening mavlink modules
        ACTIVE_MAVLINK_MODULES++;

        disableControls();
        linkQualityTimer.start(linkQualityPeriod);

        // TODO: Allow the checkbox to start and stop the logging
        if (logCheckBox->isChecked())
            mavlinkCodec.openLogFile();

        mavlinkCodec.startReading();
    }
    else
    {
        error(tr("Mavlink"), tr("Unable to open selected port."));
        startToggleButton->setState(false);
    }
}

void BaseMavlinkModule::onStopClicked()
{
    stop();
    port->close();
    enableControls();

    // Keep track of all active and listening mavlink modules
    ACTIVE_MAVLINK_MODULES--;
}

void BaseMavlinkModule::onCommandReceived(const Message &msg)
{
    qDebug() << "Command received: " << msg.getTopic().toString() << " - "
             << msg.toString();

    if (!targetSystem.has_value())
    {
        error(tr("Mavlink"), tr("No target system selected."));
        return;
    }

    auto sysId = targetSystem.value().second;

    mavlink_message_t mav_msg;
    if (!mavlinkCodec.encodeMessage(msg, sysId, MavlinkCodec::CompIdNone,
                                    mav_msg))
    {
        error(tr("Mavlink"),
              tr("Cannot encode command into a valid mavlink message."));
        return;
    }

    // If we are not active, don't process this message
    if (!startToggleButton->state())
    {
        // Only show this error if there are no other active modules
        if (ACTIVE_MAVLINK_MODULES == 0)
        {
            error(tr("Mavlink"), tr("Cannot send message: port is closed."));
        }

        return;
    }

    mavlinkCodec.send(mav_msg);

    Message confirmationMsg(
        Topic((QString)SkywardHubStrings::logCommandsTopic));

    auto nameField = Field(msg.getTopic().toString().replace(
        SkywardHubStrings::commandsTopic + "/", ""));
    confirmationMsg.setField("name", nameField);

    auto idField = Field((uint64_t)mav_msg.msgid);
    confirmationMsg.setField("message_id", idField);

    auto seqField = Field((uint64_t)mav_msg.seq);
    confirmationMsg.setField("sequence_number", seqField);

    MessageBroker::getInstance().publish(confirmationMsg);
}

void BaseMavlinkModule::onOpenLogFolderClick()
{
    QDir dir(SkywardHubStrings::defaultLogsFolder);
    if (dir.exists())
    {
        QDesktopServices::openUrl(
            QUrl::fromLocalFile(SkywardHubStrings::defaultLogsFolder));
    }
    else
    {
        warning(tr("Mavlink"), tr("The log folder does not exist yet. It will "
                                  "be created when opening the serial port."));
    }
}

void BaseMavlinkModule::enableControls()
{
    sysIdComboBox->setEnabled(true);
    logCheckBox->setEnabled(true);
    childEnableControls();
}

void BaseMavlinkModule::disableControls()
{
    sysIdComboBox->setEnabled(false);
    logCheckBox->setEnabled(false);
    childDisableControls();
}

void BaseMavlinkModule::stop()
{
    linkQualityTimer.stop();
    mavlinkCodec.stopReading();
}

void BaseMavlinkModule::setupUi()
{
    QHBoxLayout *outerLayout = new QHBoxLayout;
    outerLayout->setContentsMargins(6, 0, 6, 0);

    startToggleButton = new ToggleButton;
    startToggleButton->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(startToggleButton);
    connect(startToggleButton, &ToggleButton::toggled, this,
            &BaseMavlinkModule::onStartStreamToggled);

    QLabel *sysIdLabel = new QLabel("Target system:");
    sysIdLabel->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(sysIdLabel);

    sysIdComboBox = new QComboBox;
    sysIdComboBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));

    sysIdComboBox->addItem("Main", MavlinkCodec::SysIdMain);
    sysIdComboBox->addItem("Payload", MavlinkCodec::SysIdPayload);
    sysIdComboBox->addItem("Rig", MavlinkCodec::SysIdRig);
    sysIdComboBox->addItem("Gs Receiver", MavlinkCodec::SysIdGsReceiver);
    outerLayout->addWidget(sysIdComboBox);
    connect(sysIdComboBox, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &BaseMavlinkModule::onTargetSysChanged);
    sysIdComboBox->setCurrentIndex(0);
    onTargetSysChanged(0);

    childLayout = new QHBoxLayout;
    outerLayout->addLayout(childLayout);

    outerLayout->addStretch();

    rateLabel = new QLabel("Rate: 0 msg/s");
    rateLabel->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(rateLabel);

    logCheckBox = new QCheckBox("Enable log");
    logCheckBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(logCheckBox);

    QPushButton *openLogFolderButton = new QPushButton("Log folder");
    openLogFolderButton->setSizePolicy(
        QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    outerLayout->addWidget(openLogFolderButton);
    connect(openLogFolderButton, &QPushButton::clicked, this,
            &BaseMavlinkModule::onOpenLogFolderClick);

    setLayout(outerLayout);
}
